@@ -52,7 +52,7 @@ def pre_transform(topology: Box) -> None:
   link_transform("pre_transform",topology)
 
 """
-pre_node_transform: executed just before the node data model transformation is started
+pre/post_node_transform: executed just before/after the node data model transformation is started
 
 """
 def pre_node_transform(topology: Box) -> None:
@@ -60,8 +60,13 @@ def pre_node_transform(topology: Box) -> None:
   node_transform("pre_node_transform",topology)
   link_transform("pre_node_transform",topology)
 
+def post_node_transform(topology: Box) -> None:
+  module_transform("post_node_transform",topology)
+  node_transform("post_node_transform",topology)
+  link_transform("post_node_transform",topology)
+
 """
-pre_link_transform: executed just before the link data model transformation is started
+pre/post_link_transform: executed just before/after the link data model transformation is started
 
 """
 def pre_link_transform(topology: Box) -> None:
@@ -69,6 +74,11 @@ def pre_link_transform(topology: Box) -> None:
   node_transform("pre_link_transform",topology)
   link_transform("pre_link_transform",topology)
 
+def post_link_transform(topology: Box) -> None:
+  module_transform("post_link_transform",topology)
+  node_transform("post_link_transform",topology)
+  link_transform("post_link_transform",topology)
+
 """
 post_transform:
   execute module-specific code after the main link- and node
@@ -94,8 +104,8 @@ def augment_node_module(topology: Box) -> None:
     if not 'module' in n and n.get('role') != 'host' and devices.get_device_attribute(n,'role',topology.defaults) != 'host':
       n.module = module
 
-# Check whether the modules defined on individual nodes are valid
-# and supported
+# Check whether the modules defined on individual nodes are valid module names
+# and supported by the node device type
 #
 def check_supported_node_devices(topology: Box) -> None:
   for name,n in topology.nodes.items():
@@ -352,9 +362,9 @@ def check_module_parameters(topology: Box) -> None:
                   common.IncorrectValue,
                   'modules')
           else:
-            if type(intf[m]).__name__ != str(mod_attr[m].interface):
+            if type(intf[m]).__name__ != str(mod_attr[m].interface) and not (intf[m] is False):
               common.error(
-                f"Node {n} has invalid value type for attribute {k} for module {m} on link {l}, expected {mod_attr[m].interface}",
+                f"Node {n} has invalid value type for module {m} on link {l}, expected {mod_attr[m].interface}",
                 common.IncorrectValue,
                 'modules')
 
@@ -396,13 +406,13 @@ def check_module_dependencies(topology:  Box) -> None:
   for m in topology.get("module",[]):               # Use this format in case we don't use any modules
     mod_def = topology.defaults.get(m,{})           # Get module defaults
     if mod_def:                                     # Are they meaningful?
-      if "requires" in mod_def:                     # Do they include list of required modules?
-        for rqm in topology.defaults[m].requires:   # Loop over prerequisites
-          if not rqm in topology.module:            # Now we can be explicit - we know topology.modules exists
-            common.error(
-              f"Module {m} requires module {rqm} which is not enabled in your topology",
-              common.MissingValue,
-              'modules')
+      mod_requires = topology.defaults[m].get('requires',[])
+      for rqm in mod_requires:                      # Loop over prerequisite modules
+        if not rqm in topology.module:              # Now we can be explicit - we know topology.modules exists
+          common.error(
+            f"Module {m} requires module {rqm} which is not enabled in your topology",
+            common.MissingValue,
+            'modules')
 
       for n in topology.nodes.values():                   # Now iterate over nodes and check device-specific requirements
         if not m in n.get('module',[]):                   # Is the module we're currently checking used by this node?
@@ -410,13 +420,16 @@ def check_module_dependencies(topology:  Box) -> None:
         features = fcache.get(n.name) or \
                      devices.get_device_features(n,topology.defaults)
         fcache[n.name] = features                         # Get device features and save them in per-node cache
-        if m in features and 'requires' in features[m]:   # Check modules with device-specific requirements
-          for rqm in features[m].requires:                # ... iterate over device-specific module requirements
-            if not rqm in n.get('module'):                # ... is required module listed in the node?
-              common.error(
-                f"Module {m} on device {n.device} (node {n.name}) requires {rqm} module",
-                common.IncorrectValue,
-                'modules')
+        if m in features:                                 # If the module has device feature flags, add device requirements to global ones
+          node_requires = mod_requires + features[m].get('requires',[])
+        else:
+          node_requires = mod_requires                    # ... otherwise use global module requirements
+        for rqm in node_requires:                         # Iterate over module requirements for this node (global + device-specific)
+          if not rqm in n.get('module'):                  # ... is required module listed in the node?
+            common.error(
+              f"Module {m} on device {n.device} (node {n.name}) requires {rqm} module",
+              common.IncorrectValue,
+              'modules')
 
 """
 reorder_node_modules:
@@ -465,14 +478,26 @@ Example: copy node-level OSPF area into interfaces that don't have explicit area
 def copy_node_data_into_interfaces(topology: Box) -> None:
   for n in topology.nodes.values():
     for m in n.get('module',[]):                                 # Iterate over node modules
-      if topology.defaults[m].attributes.node_copy:              # .. any copyable attributes for this module?
-        copy_attr = Box({ k: v
-          for k,v in n.get(m,{}).items()
-            if k in topology.defaults[m].attributes.node_copy }) # Build a Box of node attributes that could be copied to interfaces
-        if copy_attr:                                            # .. anything to copy?
-          for intf in n.get('interfaces',[]):                    # .. if so, it would be nice to merge it with interface data
-            if isinstance(intf.get(m,{}),dict):                  # .. but only if the interface data allows it
-              intf[m] = copy_attr + intf[m]
+      mod_attr = topology.defaults[m].attributes                 # ... get a pointer to module attributes to make code easier to read
+      if not mod_attr.node_copy:                                 # Any copyable attributes for this module?
+        continue                                                 # ... nope, get out of here
+
+      copy_attr = Box({ k: v
+        for k,v in n.get(m,{}).items()
+          if k in mod_attr.node_copy })                          # Build a Box of node attributes that could be copied to interfaces
+
+      for intf in n.get('interfaces',[]):                        # We might have some work to do, iterate over all interfaces
+        if not isinstance(intf.get(m,{}),dict):                  # ... if the interface module data is not a dict, we can't merge
+          continue
+        if 'vrf' in intf and mod_attr.vrf_aware:                 # Do we have to deal with VRF-aware attributes?
+          vrf_attr = Box({ k: v
+            for k,v in n.vrfs[intf.vrf].get(m,{}).items()
+              if k in mod_attr.vrf_aware })                      # Build a Box of VRF attributes that could be copied to interfaces
+        else:
+          vrf_attr = Box({})                                     # ... otherwise set VRF attributes to an empty box
+
+        if copy_attr or vrf_attr:                                # ... modify interface data only if we have something to merge
+          intf[m] = copy_attr + vrf_attr + intf[m]
 
 """
 get_effective_module_attribute:
@@ -519,6 +544,9 @@ mod_load: typing.Dict = {}
 def module_transform(method: str, topology: Box) -> None:
   global mod_load
 
+  if common.debug_active('modules'):
+    print(f'Processing module_{method} hooks')
+
   for m in topology.get('module',[]):
     if not mod_load.get(m):
       mod_load[m] = _Module.load(m,topology.get(m))
@@ -530,6 +558,9 @@ def module_transform(method: str, topology: Box) -> None:
 def node_transform(method: str , topology: Box) -> None:
   global mod_load
 
+  if common.debug_active('modules'):
+    print(f'Processing node_{method} hooks')
+
   for name,n in topology.nodes.items():
     for m in n.get('module',[]):
       if not mod_load.get(m):  # pragma: no cover (module should have been loaded already)
@@ -542,6 +573,9 @@ def node_transform(method: str , topology: Box) -> None:
 def link_transform(method: str, topology: Box) -> None:
   global mod_load
 
+  if common.debug_active('modules'):
+    print(f'Processing link_{method} hooks')
+
   for l in topology.get("links",[]):
     mod_list: typing.Dict = {}
     for node_data in l.get(IFATTR,[]):
