@@ -7,17 +7,12 @@ from box import Box
 from . import _Module,_routing,get_effective_module_attribute
 from .. import common
 from .. import data
+from ..data import global_vars
 from .. import addressing
 from ..data import get_from_box,get_global_parameter
-from ..augment import devices
+from ..augment import devices,groups
 from ..augment import links
 
-# Global variables -- would love to be without them, but the alternatives
-# are even messier
-#
-vlan_ids: Box
-vlan_next: Box
-
 # Static lists of keywords
 #
 vlan_mode_kwd: typing.Final[list] = [ 'bridge', 'irb', 'route' ]
@@ -33,14 +28,6 @@ phy_ifattr: typing.Final[list] = ['bridge','ifindex','parentindex','ifname','lin
 keep_subif_attr: typing.Final[list] = ['vlan','ifindex','ifname','type']    # Keep these attributes on VLAN subinterfaces
 vlan_link_attr_copy: typing.Final[list] = ['role','unnumbered','pool']      # VLAN attributes to copy to member links
 
-"""
-init_global_vars: Initialize the VLAN ID pool
-"""
-def init_global_vars() -> None:
-  global vlan_ids, vlan_next
-  vlan_ids = Box({},default_box=True,box_dots=True)
-  vlan_next = Box({},default_box=True,box_dots=True)
-
 #
 # build_vlan_id_set: given an object (topology or node), create a set of VLAN attributes (ID or VNI)
 # that appear in that object.
@@ -51,11 +38,12 @@ def build_vlan_id_set(obj: Box, attr: str, objname: str) -> set:
       common.fatal(f'Found a "vlans" setting that is not a dictionary in {objname}','vlan')
       return set()
 
-    return { v[attr] for v in obj.vlans.values() if isinstance(v,dict) and attr in v and v[attr] is not None }
+    return { v[attr] for v in obj.vlans.values() if isinstance(v,dict) and attr in v and v[attr] is not None and not isinstance(v[attr],bool) }
   return set()
 
 def populate_vlan_id_set(topology: Box) -> None:
-  global vlan_ids, vlan_next
+  vlan_ids = global_vars.get('vlan_ids')
+  vlan_next = global_vars.get('vlan_next')
 
   for k in ('id','vni'):
     vlan_ids[k] = build_vlan_id_set(topology,k,'topology')
@@ -73,7 +61,8 @@ def populate_vlan_id_set(topology: Box) -> None:
         vlan_ids[k] = vlan_ids[k].union(build_vlan_id_set(n,k,n.name))
 
 def get_next_vlan_id(k : str) -> int:
-  global vlan_ids,vlan_next
+  vlan_ids = global_vars.get('vlan_ids')
+  vlan_next = global_vars.get('vlan_next')
 
   if k not in vlan_ids or k not in vlan_next:   # pragma: no cover
     common.fatal(f'Invalid attribute {k} in get_next_vlan_id call')
@@ -84,6 +73,15 @@ def get_next_vlan_id(k : str) -> int:
   vlan_ids[k].add(vlan_next[k])
   return vlan_next[k]
 
+#
+# Define a utility function to check whether a VLAN ID or VNI is used to hide internal data structures
+# from modules needing this functionality
+#
+def is_vlan_id_used(vlan_id: int, namespace: str) -> bool:
+  vlan_ids = global_vars.get('vlan_ids')
+
+  return vlan_id in vlan_ids[namespace]
+
 #
 # routed_access_vlan: Given a link with access/native VLAN, check if all nodes on the link use routed VLAN
 #
@@ -131,7 +129,7 @@ def interface_vlan_mode(intf: Box, node: Box, topology: Box) -> str:
 # * VLAN forwarding mode
 #
 def validate_vlan_attributes(obj: Box, topology: Box) -> None:
-  global vlan_ids
+  vlan_ids = global_vars.get('vlan_ids')
 
   obj_name = 'global VLANs' if obj is topology else obj.name
   obj_path = 'vlans' if obj is topology else f'nodes.{obj.name}.vlans'
@@ -174,19 +172,28 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
       common.error(f'VLAN ID {vdata.id} for VLAN {vname} in {obj_name} must be between 2 and 4094',common.IncorrectValue,'vlan')
       continue
 
-    if not 'vni' in vdata:                                          # When VNI is not defined
+    # Assign VNI only when topology.defaults.vlan.auto_vni is set and there's no VNI in VLAN data
+    #
+    assign_vni = topology.defaults.vlan.auto_vni and not 'vni' in vdata
+    if 'vni' in vdata:
+      if vdata.vni is True:                                         # VNI auto-assignment requested
+        assign_vni = True
+      elif vdata.vni is False:                                      # Explicit request not to assign a VNI
+        vdata.pop('vni')
+      elif not isinstance(vdata.vni,int):                           # VNI is not bool, validate that it's int and in correct range
+        common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
+        continue
+      elif vdata.vni < 2 or vdata.vni > 16777215:
+        common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be between 2 and 16777215',common.IncorrectValue,'vlan')
+        continue
+
+    if assign_vni:                                                  # So far we either have a valid VNI or know whether to auto-assign it
       vni_default = topology.defaults.vlan.start_vni + vdata.id     # ... try to build VNI from VLAN ID
-      if not vni_default in vlan_ids.vni:                           # Is the VNI free?
+      if not is_vlan_id_used(vni_default,'vni'):                    # Is the VNI free?
         vdata.vni = vni_default                                     # ... great, take it
         vlan_ids.vni.add(vni_default)                               # ... and add it to the list of used VNIs
       else:                                                         # Too bad, we had such a great idea but it failed
         vdata.vni = get_next_vlan_id('vni')                         # ... so take the next available VNI
-    if not isinstance(vdata.vni,int):                               # Not done yet, we still have to validate the VNI type and range
-      common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
-      continue
-    if vdata.vni < 2 or vdata.vni > 16777215:
-      common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be between 2 and 16777215',common.IncorrectValue,'vlan')
-      continue
 
     vlan_pool = [ vdata.pool ] if 'pool' in vdata else []
     vlan_pool.extend(['vlan','lan'])
@@ -359,31 +366,6 @@ def validate_link_vlan_attributes(link: Box,v_attr: Box,topology: Box) -> bool:
                 intf[af] = False                                  # ... make sure it's not connected to the native VLAN subnet
   return link_ok
 
-"""
-check_native_routed_vlan: verify that all devices attached to a link can support native routed VLAN
-
-This function is called when we're dealing with a link with routed non-tagged VLAN. It needs to check
-whether we're dealing with a native VLAN (as opposed to access VLAN) and if so, iterate through
-the attached nodes to validate whether they support this setup.
-"""
-
-def check_native_routed_vlan(link: Box, v_attr: Box, topology: Box) -> None:
-  if not 'native' in v_attr:                                      # No vlan.native attribute, we're good
-    return
-
-  if not v_attr.native.set:                                       # Native VLAN set is empty, we're good
-    return
-
-  native_vlan = list(v_attr.native.set)[0]
-  for intf in link.interfaces:                                    # Check only nodes VLAN attributes that have native VLAN in their trunk
-    if 'vlan' in intf and native_vlan in intf.vlan.get('trunk',{}):
-      features = devices.get_device_features(topology.nodes[intf.node],topology.defaults)
-      if not features.vlan.native_routed:
-        common.error(
-          f'Node {intf.node} does not support routed native VLANs\n... link {link}',
-          common.IncorrectValue,
-          'vlan')
-
 """
 copy_vlan_attributes: copy prefix and link type from vlan to link
 """
@@ -600,7 +582,8 @@ def create_svi_interfaces(node: Box, topology: Box) -> dict:
   iflist_len = len(node.interfaces)
   for ifidx in range(0,iflist_len):
     ifdata = node.interfaces[ifidx]
-    access_vlan = get_from_box(ifdata,'vlan.access') or get_from_box(ifdata,'vlan.native')
+    native_vlan = get_from_box(ifdata,'vlan.native')
+    access_vlan = get_from_box(ifdata,'vlan.access') or native_vlan
     if not access_vlan:                                                     # No access VLAN on this interface?
       continue                                                              # ... good, move on
 
@@ -626,6 +609,10 @@ def create_svi_interfaces(node: Box, topology: Box) -> dict:
       if vlan_subif:
         ifdata.vlan.mode = 'route'
         ifdata.vlan.access_id = vlan_data.id
+      else:
+        ifdata._vlan_mode = 'route'                                         # Flags we need to clean up up the routed native VLAN mess
+        if native_vlan:                                                     # ... we can't use the vlan.* attributes as they might get removed
+          ifdata._vlan_native = native_vlan
 
       node.interfaces[ifidx] = vlan_copy + ifdata                           # Merge VLAN data with interface data
       continue                                                              # Move to next interface
@@ -750,6 +737,17 @@ def rename_neighbor_interface(node: Box, neighbor_name: str, old_ifname: str, ne
         if n.node == neighbor_name and n.ifname == old_ifname:
           n.ifname = new_ifname
 
+"""
+Utility function: remove_vlan_from_trunk -- we need it to unify routed VLANs and routed native VLAN cases
+"""
+def remove_vlan_from_trunk(parent_intf: Box, vlan_id: int) -> None:
+  parent_intf.vlan.trunk_id = [ vlan for vlan in parent_intf.vlan.trunk_id if vlan != vlan_id ]
+  if not parent_intf.vlan.trunk_id:                               # Nothing left?
+    parent_intf.vlan.pop('trunk_id',None)                         # ... remove all trunk-related attributes from parent interface
+    parent_intf.vlan.pop('trunk',None)
+    if not parent_intf.vlan:                                      # And remove the VLAN dictionary if it's no longer needed
+      parent_intf.pop('vlan',None)
+
 """
 rename_vlan_subinterfaces: rename or remove interfaces created from VLAN pseudo-links
 """
@@ -817,18 +815,47 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
         intf.pop(attr,None)
 
     if intf.vlan.access_id in parent_intf.vlan.get('trunk_id',[]):    # Remove subinterface VLAN from parent interface trunk list
-      parent_intf.vlan.trunk_id = [ vlan for vlan in parent_intf.vlan.trunk_id if vlan != intf.vlan.access_id ]
-      if not parent_intf.vlan.trunk_id:                               # Nothing left?
-        parent_intf.vlan.pop('trunk_id',None)                         # ... remove all trunk-related attributes from parent interface
-        parent_intf.vlan.pop('trunk',None)
-        if not parent_intf.vlan:                                      # And remove the VLAN dictionary if it's no longer needed
-          parent_intf.pop('vlan',None)
+      remove_vlan_from_trunk(parent_intf,intf.vlan.access_id)
 
-  # Final check: mixed bridged/routed trunks
-  #
-  # We can skip this check if the device supports mixed routed/bridged trunks or if it
-  # uses VLAN subinterfaces to implement VLANs.
-  #
+"""
+If we have a routed native VLAN, and there are no other bridged VLANs left
+in the VLAN trunk, we don't need a trunk at all.
+
+We are pretty late in the process, so we have to identify a native VLAN in
+a roundabout way: vlan_name is set to the native VLAN name, and there's still
+vlan.trunk_id on the link.
+
+The check for routed VLAN is also roundabout: if we have an IPv4 or IPv6 address
+on the interface, then we can assume it's a routed VLAN.
+"""
+def cleanup_routed_native_vlan(node: Box, topology: Box) -> None:
+  for intf in node.interfaces:
+    if not '_vlan_native' in intf:                                  # Skip interfaces without native VLAN
+      continue
+
+    if intf._vlan_mode != 'route':                                  # Nothing to do if the native VLAN is not routed
+      continue
+
+    features = devices.get_device_features(node,topology.defaults)
+    if not features.vlan.native_routed:
+      common.error(
+        f'Node {node.name} device {node.device} does not support native routed VLANs (vlan {intf._vlan_native} interface {intf.name})',
+        common.IncorrectValue,
+        'vlan')
+
+    vlan_id = node.vlans.get(intf._vlan_native,{}).id
+    if vlan_id:
+      remove_vlan_from_trunk(intf,vlan_id)
+
+"""
+After cleaning up the VLAN trunks as much as possible, it's time to check
+whether we're dealing with a mixed bridged/routed trunk.
+
+We can skip this check if the device supports mixed routed/bridged trunks or if it
+uses VLAN subinterfaces to implement VLANs.
+"""
+def check_mixed_trunks(node: Box, topology: Box) -> None:
+  features = devices.get_device_features(node,topology.defaults)
   if features.vlan.mixed_trunk:
     return
 
@@ -885,14 +912,28 @@ def fix_vlan_gateways(topology: Box) -> None:
     for intf in node.get('interfaces',[]):                            # Iterate over all interfaces
       if not get_from_box(intf,'gateway.ipv4'):                       # ... that don't have an IPv4 gateway
         for neighbor in intf.get('neighbors',[]):                     # Iterate over all neighbors
-          if 'ipv4' in neighbor:                                      # ... until we find one with a usable IPv4
-            intf.gateway.ipv4 = neighbor.ipv4                         # Set that address as our gateway
-            break                                                     # ... and get out of here
+          if neighbor.get('ipv4',False):                              # ... until we find one with a usable IPv4
+            n_node = topology.nodes[neighbor.node]
+            if n_node.get('role') != 'host':                          # ... that is not another host
+              intf.gateway.ipv4 = neighbor.ipv4                       # Set that address as our gateway
+              break                                                   # ... and get out of here
+
+"""
+populate_node_vlan_data -- merge topology VLANs into node VLANs that were copied from groups.node_data
+"""
+def populate_node_vlan_data(n: Box, topology: Box) -> None:
+  if 'vlans' in n:                                                          # Copy topology VLAN data into node VLAN data
+    for vname in n.vlans.keys():                                            # ... to cope with nodes that had VLANs defined
+      if vname in topology.get('vlans',{}):                                 # ... through groups.node_data
+        topo_data = Box(topology.vlans[vname])                              # Create a copy of topology VLAN
+        topo_data.pop('neighbors',None)                                     # ... and remove neighbors
+        n.vlans[vname] = topo_data + n.vlans[vname]                         # ... now merge with the VLAN data
 
 class VLAN(_Module):
 
   def module_pre_transform(self, topology: Box) -> None:
-    init_global_vars()
+    if 'groups' in topology:
+      groups.export_group_node_data(topology,'vlans','vlan',copy_keys=['id','vni'])
     if get_from_box(topology,'vlan.mode'):
       if topology.vlan.mode not in vlan_mode_kwd:     # pragma: no cover
         common.error(
@@ -967,7 +1008,6 @@ class VLAN(_Module):
             link[k] = vlan_data[k] + link[k]
 
     if routed_vlan:
-      check_native_routed_vlan(link,v_attr,topology)
       link.vlan.mode = 'route'
       for intf in link.interfaces:
         if 'vlan' in intf:
@@ -984,12 +1024,15 @@ class VLAN(_Module):
             intf.ipv4 = False                                                     # ... if so, disable addressing on this interface
             intf.ipv6 = False
 
-  def module_post_transform(self, topology: Box) -> None:
+  def module_post_link_transform(self, topology: Box) -> None:
     for n in topology.nodes.values():
       if 'vlan' in n.get('module',[]):
+        populate_node_vlan_data(n,topology)
         vlan_ifmap = create_svi_interfaces(n,topology)
         map_trunk_vlans(n,topology)
         rename_vlan_subinterfaces(n,topology)
+        cleanup_routed_native_vlan(n,topology)
+        check_mixed_trunks(n,topology)
 
     for n in topology.nodes.values():
       set_svi_neighbor_list(n,topology)
